#ifndef MYX_REDIS_CLIENT_INL_HPP_
#define MYX_REDIS_CLIENT_INL_HPP_

#pragma once

#ifndef MYXLIB_HEADER_ONLY
#include <myx/redis/client.hpp>
#endif

#include <myx/redis/containers.hpp>

#include <QThread>
#include <QTcpSocket>
#include <QElapsedTimer>
#include <QJsonDocument>
#include <QByteArrayList>
#include <QCoreApplication>


namespace myx {

namespace redis {

namespace constants {

extern const QByteArray Separator; // { "\r\n" }; //!< Строковый разделитель слов во фразе

} // namespace constants

namespace C = constants;

/*!
 * \brief Создает объект класса
 * \param parent Родительский объект
 * \param database Идентификатор БД
 */
Client::Client( int database, QObject* parent ) :
	Base{ parent }
{
	_database = database;
}


/*!
 * \brief Удаляет все созданные соединения
 */
Client::~Client()
{
	for ( auto iter: _session )
	{
		if ( iter->state() == QTcpSocket::ConnectedState )
		{
			iter->disconnectFromHost();
			if ( iter->state() != QTcpSocket::UnconnectedState )
			{
				iter->waitForDisconnected();
			}
		}
		iter->deleteLater();
	}
}


/*!
 * \brief Возвращает идентификатор БД
 * \return Идентификатор БД
 */
int Client::database() const
{
	return( _database );
}


/*!
 * \brief Выполняет переключение текущей БД
 * \param database Номер новой БД, к которой будет организовано подключение
 */
void Client::select( int database )
{
	if ( _database == database )
	{
		return;
	}
	_database      = database;
	_outOfDatabase = { _session.begin(), _session.end() };
}


/*!
 * \brief Определяет сколько из переданных ключей существует в redis
 * \param keys Rk.xb
 * \return Количество существующих ключей
 */
int Client::exists( const QStringList& keys )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( 0 );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( 0 );
	}
	socket->write( array( QVariantList { "EXISTS" } + QVariant( keys ).toList() ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( 0 );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) )
	{
		QRegExp rx { "\\d+" };
		rx.indexIn( buffer );
		return( rx.cap().toInt() );
	}
	return( -1 );
} // Client::exists


/*!
 * \brief Очищает БД
 */
bool Client::flushdb()
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "$7", "FLUSHDB" } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::flushdb


/*!
 * \brief Удаляет данные по списку ключей
 * \param keys Список ключей для удаления
 * \return Признак успешной операции
 */
bool Client::del( const QStringList& keys )
{
	if ( keys.empty() )
	{
		return( true );
	}
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "DEL" } + QVariant( keys ).toList() ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::del


/*!
 * \brief Устанавливает таймаут на значение в секундах
 * \param key Ключ значения
 * \param secs Таймаут в секундах
 * \return Признак успешной установки
 */
bool Client::expire( const QString& key, int secs )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "EXPIRE", key, secs } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( true );
} // Client::expire


/*!
 * \brief Устанавливает таймаут на значение в милисекундах
 * \param key Ключ значения
 * \param secs Таймаут в милисекундах
 * \return Признак успешной установки
 */
bool Client::pexpire( const QString& key, int msecs )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "PEXPIRE", key, msecs } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( true );
} // Client::pexpire


/*!
 * \brief Удаляет таймаут со значения, делаю его persistent
 * \param key Ключ значения
 * \return Признак успешного удаления таймаута
 */
bool Client::persist( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "PERSIST", key } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( true );
} // Client::persist


/*!
 * \brief Публикует сообщение
 * \param channel Канал сообщений
 * \param message Текст сообщений
 * \return Признак удачной публикации
 */
bool Client::publish( const QString& channel, const QByteArray& message )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	if ( message.isEmpty() )
	{
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "PUBLISH", channel, message } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( true );
} // Client::publish


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param channel Канал сообщений
 * \param message Текст сообщений
 * \return Признак удачной публикации
 */
bool Client::publish( const QString& channel, const QByteArray& message,
                      const QString& host, quint16 port )
{
	if ( message.isEmpty() )
	{
		return( false );
	}
	QTcpSocket socket;
	do {
		if ( socket.state() == QAbstractSocket::UnconnectedState )
		{
			socket.connectToHost( host, port );
		}
		QCoreApplication::processEvents();
	} while( socket.state() != QAbstractSocket::ConnectedState );
	socket.write( array( { "PUBLISH", channel, message } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket.state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( true );
} // Client::publish


/*!
 * \brief Выполняет команду SET протокола RESP
 * \param key Ключ-строка
 * \param value Значение по ключу
 * \return Признак удачной установки значения
 */
bool Client::set( const QString& key, const QVariant& value )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "SET", key, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::set


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ данных
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Строка данных
 */
bool Client::set( const QString& key, const QVariant& value,
                  int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "SET", key, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket.state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::set


/*!
 * \brief Загружает lua скрипт на сервер
 * \param script Код скрипта
 * \return SHA загруженного скрипта (пустая SHA в случае возникновения ошибки)
 */
QByteArray Client::scriptLoad( const QByteArray& script )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "SCRIPT", "LOAD", script } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer.startsWith( "-ERR" ) )
			{
				qWarning() << "loading redis script error:" << buffer;
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				return( parts.front() );
			}
			return {};
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::scriptLoad


/*!
 * \brief Синхронный метод получения значения
 * \param key Ключ
 * \return Строка данных
 */
QByteArray Client::get( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "GET", key } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer.startsWith( "-WRONG" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				return( parts.front() );
			}
			return {};
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::get


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ данных
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Строка данных
 */
QByteArray Client::get( const QString& key, int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "GET", key } ) );
	QByteArray buffer;
	QByteArray value;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				value = parts.front();
			}
			break;
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( value );
} // Client::get


/*!
 * \brief Задает значение, хранящееся в хэш-таблице
 * \param key Ключ таблицы
 * \param field Ключ значения
 * \param value Значение
 * \return Признак успешной операции
 */
bool Client::hset( const QString& key, const QString& field, const QVariant& value )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "HSET", key, field, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::hset


/*!
 * \brief Заполняет hash по ключу
 * \param key Ключ
 * \param hash Hash таблица со значениями
 * \return Признак корректного заполения
 */
bool Client::hmset( const QString& key, const QVariantHash& hash )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	QVariantList parts { "HMSET", key };
	for ( auto iter = hash.cbegin(); iter != hash.cend(); iter++ ) parts << iter.key() << iter.value();
	socket->write( array( parts ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::hmset


/*!
 * \brief Заполняет удаляет данные из хэш-таблицы redis
 * \param key Ключ таблицы для redis
 * \param hashKeys Ключи удаляемых данных
 * \return Признак корректного заполения
 */
bool Client::hdel( const QString& redisKey, const QStringList& hashKeys )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "HDEL", redisKey } + QVariant( hashKeys ).toList() ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::hdel


/*!
 * \brief Заполняет hash по ключу
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ
 * \param hash Hash таблица со значениями
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Признак корректного заполения
 */
bool Client::hmset( const QString& key, const QVariantHash& hash, int database,
                    const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	QVariantList parts { "HMSET", key };
	for ( auto iter = hash.cbegin(); iter != hash.cend(); iter++ ) parts << iter.key() << iter.value();
	socket.write( array( parts ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket.state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( buffer == "+OK\r\n" );
} // Client::hmset


/*!
 * \brief Синхронный метод хэш-таблицы
 * \param key Ключ
 * \return Таблица данных
 */
QVariantHash Client::hgetall( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "HGETALL", key } ) );
	QByteArray buffer;
	QVariantHash hash;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer.startsWith( "-WRONG" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			for ( int i = 1; i < parts.size(); i += 2 ) hash[parts[i - 1]] = parts[i];
			return( hash );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::hgetall


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ данных
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Таблица данных
 */
QVariantHash Client::hgetall( const QString& key, int database,
                              const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "HGETALL", key } ) );
	QByteArray buffer;
	QVariantHash hash;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) )
			{
				break;
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			for ( int i = 1; i < parts.size(); i += 2 ) hash[parts[i - 1]] = parts[i];
			break;
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( hash );
} // Client::hgetall


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 * \param redisKey Ключ данных для redis
 * \param hashKey Ключ внутри запрашиваемой таблицы
 * \return Значение из таблицы
 */
QByteArray Client::hget( const QString& redisKey, const QString& hashKey )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "HGET", redisKey, hashKey } ) );
	QByteArray value;
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				return( parts.front() );
			}
			return {};
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::hget


/*!
 * \brief Получает данные из redis без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param redisKey Ключ данных для redis
 * \param hashKey Ключ внутри запрашиваемой таблицы
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Значение из таблицы
 */
QByteArray Client::hget( const QString& redisKey, const QString& hashKey,
                         int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "HGET", redisKey, hashKey } ) );
	QByteArray value;
	QByteArray buffer;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				value = parts.front();
			}
			break;
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( value );
} // Client::hget


/*!
 * \brief Добавляет элементы в конец списка
 * \param key Ключ списка
 * \param values Добавляемые элементы
 * \return Признак успешной операции
 */
bool Client::rpush( const QString& key, const QVariantList& values )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "RPUSH", key } + values ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::rpush


/*!
 * \brief Добавляет элементы в начало списка
 * \param key Ключ списка
 * \param values Добавляемые элементы
 * \return Признак успешной операции
 */
bool Client::lpush( const QString& key, const QVariantList& values )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "LPUSH", key } + values ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::lpush


/*!
 * \brief Добавляет задает значение элемента по его индексу
 * \param key Ключ списка
 * \param index Индекс элемента
 * \param index Значение элемента
 * \return Признак успешной операции
 */
bool Client::lset( const QString& key, int index, const QVariant& value )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	auto indexString = QByteArray::number( index );
	auto valueString = value.toString();
	socket->write( array( { "LSET", key, index, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::lset


/*!
 * \brief Оставляет только элементы списка внутри диапазона from...to
 * \param key Ключ списка
 * \param from Начало диапазона (0-based)
 * > 0: от начала списка
 * < 0: от конца списка
 * \param to Конец диапазона (0-based)
 * > 0: от начала списка
 * < 0: от конца списка
 * \return Признак успешной операции
 */
bool Client::ltrim( const QString& key, int from, int to )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "LTRIM", key, from, to } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == "+OK\r\n" );
} // Client::ltrim


/*!
 * \brief Удаляет count первых элементов списка
 * \param key Ключ списка
 * \param count Количество удаляемых элементов
 * count > 0 от начала списка
 * count < 0 от конца списка
 * count = 0 удаляет все элементы
 * \param value Значение удаляемых элементов
 * \return Признак успешной операции
 */
bool Client::lrem( const QString& key, const QVariant& value, int count )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "LREM", key, count, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::lrem


/*!
 * \brief Возвращает длину списка
 * \param key Ключ списка
 * \return Длина списка или -1 в случае ошибки операции
 */
int Client::llen( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( -1 );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( -1 );
	}
	socket->write( array( { "LLEN", key } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( -1 );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) )
	{
		QRegExp rx { "\\d+" };
		rx.indexIn( buffer );
		return( rx.cap().toInt() );
	}
	return( -1 );
} // Client::llen


/*!
 * \brief Добавляет элементы в начало сета
 *
 * O(N), N - количество добавляемых элементов
 * \param key Ключ сета
 * \param values Добавляемые элементы
 * \return Признак успешной операции
 */
bool Client::sadd( const QString& key, const QVariantList& values )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "SADD", key } + values ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::sadd


/*!
 * \brief Удаляет значения из сета
 *
 * O(N), N - количество добавляемых элементов
 * \param key Ключ сета
 * \param values Добавляемые элементы
 * \return Признак успешной операции
 */
bool Client::srem( const QString& key, const QVariantList& values )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( QVariantList { "SREM", key } + values ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) );
} // Client::srem


/*!
 * \brief Проверяет является ли значение частью набора
 * \param key Ключ набора
 * \param value Проверяемое значение
 * \return Признак того, что значение входит в набор
 */
bool Client::sismember( const QString& key, const QVariant& value )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( false );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( false );
	}
	socket->write( array( { "SISMEMBER", key, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( buffer == ":1\r\n" );
} // Client::sismember


/*!
 * \brief Проверяет является ли значение частью набора без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ набора
 * \param value Проверяемое значение
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Признак того, что значение входит в набор
 */
bool Client::sismember( const QString& key, const QVariant& value,
                        int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "SISMEMBER", key, value } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket.state() != QAbstractSocket::ConnectedState )
		{
			return( false );
		}
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( buffer == ":1\r\n" );
} // Client::sismember


/*!
 * \brief Возвращает количество элементов в наборе
 * \param key Ключ набора
 * \return КОличество элементов
 */
int Client::scard( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return( -1 );
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return( -1 );
	}
	socket->write( array( { "SCARD", key } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return( -1 );
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( QRegExp( ":\\d+\\r\\n" ).exactMatch( buffer ) )
	{
		QRegExp rx { "\\d+" };
		rx.indexIn( buffer );
		return( rx.cap().toInt() );
	}
	return( -1 );
} // Client::scard


/*!
 * \brief Возвращает элементы сета
 * \param key Ключ сета
 * \return Строковые значения сета
 */
QVariantList Client::smembers( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "SMEMBERS", key } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			QVariantList list;
			std::copy( parts.begin(), parts.end(), std::back_inserter( list ) );
			return( list );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::smembers


/*!
 * \brief Удаляет и возвращает последний элемент списка
 * \param key Ключ списка
 * \return Последний элемент
 */
QByteArray Client::rpop( const QString& key )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "RPOP", key } ) );
	QByteArray buffer;
	while ( !buffer.endsWith( C::Separator ) )
	{
		if ( socket->state() != QAbstractSocket::ConnectedState )
		{
			return {};
		}
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			return( parts.size() == 1 ? parts.front() : "" );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( "" );
} // Client::rpop


/*!
 * \brief Возвращает элемент по его позиции
 * \param key Ключ списка
 * \param index Позиция элемента
 * \return Текстовое представление элемента
 */
QByteArray Client::lindex( const QString& key, int index )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "LINDEX", key, index } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer == "$-1" + C::Separator )
			{
				return {};
			}
			buffer.append( socket->readAll() );
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			if ( parts.size() == 1 )
			{
				return( parts.front() );
			}
			return {};
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::lindex


/*!
 * \brief Реализует метод redis KEYS
 * \param pattern Шпблон для поиска ключей
 * \return Список ключей, соответствующих шиблону
 */
QStringList Client::keys( const QString& pattern )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( pattern.isEmpty() ? QVariantList { "KEYS" } :
	                      QVariantList { "KEYS", pattern } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) || buffer.startsWith( "-WRONG" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			QStringList list;
			std::copy( parts.begin(), parts.end(), std::back_inserter( list ) );
			return( list );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::keys


/*!
 * \brief Проверяет является ли значение частью набора без создания постоянного подключения
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param pattern Шаблон ключей
 * \param host Хост для установки соединения с сервером redis
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Признак того, что значение входит в набор
 */
QStringList Client::keys( const QString& pattern, int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( pattern.isEmpty() ? QVariantList { "KEYS" } :
	                     QVariantList { "KEYS", pattern } ) );
	QByteArray buffer;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) || buffer.startsWith( "-WRONG" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			QStringList list;
			std::copy( parts.begin(), parts.end(), std::back_inserter( list ) );
			return( list );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::keys


/*!
 * \brief Возвращает элементы хранимого списка
 * \param key Ключ списка
 * \param from Начальный индекс
 * \param to Конечный индекс
 * \return Строковые значения списка
 */
QVariantList Client::lrange( const QString& key, int from, int to )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( { "LRANGE", key, from, to } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) || buffer.startsWith( "-WRONG" ) )
			{
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			QVariantList list;
			std::copy( parts.begin(), parts.end(), std::back_inserter( list ) );
			return( list );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::lrange


/*!
 * \brief Извлекает из базы вектор и преобразует его к типу вектор json
 *
 * Специализация
 * \param key Ключ для измлечения данных
 * \param database Идентификатор БД
 * \param port Номер порта redis
 * \param from Начало интервала
 * \param to Конец интервала
 * \param Хост redis
 * \return Вектор значений
 */
template<>
std::vector< QJsonObject > Client::lrange< QJsonObject >( const QString& key, int from, int to,
                                                          int database,
                                                          const QString& host, quint16 port )
{
	auto list = lrange( key, from, to, database, host, port );
	std::vector< QJsonObject > data;
// TODO
//	std::transform( list.begin(), list.end(), std::back_inserter( data ),
//	                []( const auto& iter ) { return( QJsonDocument::fromJson( iter ).object() ); } );
	return( data );
}


/*!
 * \brief Возвращает элементы хранимого списка
 *
 * Метод отличается более низкой производительностью по сравнению с нестатической реализацией
 * \param key Ключ списка
 * \param from Начальный индекс
 * \param to Конечный индекс
 * \param port Порт для установки соединения с сервером redis
 * \param database Идентификатор БД
 * \return Строковые значения списка
 */
QByteArrayList Client::lrange( const QString& key, int from, int to, int database,
                               const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( { "LRANGE", key, from, to } ) );
	QByteArray buffer;
	QByteArrayList data;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( ( buffer == "$-1\r\n" ) || ( buffer == "*0\r\n" ) || buffer.startsWith( "-WRONG" ) )
			{
				break;
			}
			int splitPos;
			data = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			break;
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	if ( socket.state() == QAbstractSocket::ConnectedState )
	{
		socket.disconnectFromHost();
	}
	while ( socket.state() != QAbstractSocket::UnconnectedState ) QCoreApplication::processEvents();
	return( data );
} // Client::lrange


/*!
 * \brief Переключает сокет на работу с определенной БД
 * \param socket Сокет
 * \param database Номер БД
 * \return Признак успешной операции
 */
bool Client::select( QTcpSocket* socket, int database )
{
	socket->write( array( { "SELECT", database } ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			return( buffer == "+OK\r\n" );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return( false );
} // Client::select


/*!
 * \brief Настраивает сокет для подключения к БД
 * \param socket Сокет TCP
 * \param database База днных
 * \param host Адрес хоста redis
 * \param port Номер порта redis
 * \return Признак удачной настройки
 */
bool Client::setupSocket( QTcpSocket* socket, int database, const QString& host, quint16 port )
{
	static constexpr int timeout { 3000 };
	QElapsedTimer timer;
	timer.start();
	do {
		if ( socket->state() == QAbstractSocket::UnconnectedState )
		{
			socket->connectToHost( host, port );
		}
		QCoreApplication::processEvents();
	} while( socket->state() != QAbstractSocket::ConnectedState && timer.elapsed() < timeout );
	if ( socket->state() != QAbstractSocket::ConnectedState )
	{
		return( false );
	}
	if ( ( database != 0 ) && !select( socket, database ) )
	{
		return( false );
	}
	return( true );
} // Client::setupSocket


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
QByteArrayList Client::evalSha( const QByteArray& sha, const QVariantList& keys,
                                const QVariantList& argv )
{
	if ( QThread::currentThread() != thread() )
	{
		qDebug() << "no redis connection for caller thread";
		return {};
	}
	Session instance { _sessionCount };
	auto socket = session( rwTimeout() );
	if ( socket == nullptr )
	{
		return {};
	}
	socket->write( array( QVariantList { "EVALSHA", sha, keys.size() } + keys + argv ) );
	QByteArray buffer;
	while ( socket->state() == QAbstractSocket::ConnectedState )
	{
		if ( socket->bytesAvailable() > 0 )
		{
			buffer.append( socket->readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer.startsWith( "-ERR" ) || buffer.startsWith( "-NOSCRIPT" ) )
			{
				qWarning() << "evaluating redis script error:" << buffer;
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			return( parts );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::evalSha


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Статический вариант
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \param database Идентификатор БД
 * \param host Адрес сервера redis
 * \param port Номер порта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
QByteArrayList Client::evalSha( const QByteArray& sha, const QVariantList& keys,
                                const QVariantList& argv,
                                int database, const QString& host, quint16 port )
{
	QTcpSocket socket;
	if ( !setupSocket( &socket, database, host, port ) )
	{
		return {};
	}
	socket.write( array( QVariantList { "EVALSHA", sha, keys.size() } + keys + argv ) );
	QByteArray buffer;
	while ( socket.state() == QAbstractSocket::ConnectedState )
	{
		if ( socket.bytesAvailable() > 0 )
		{
			buffer.append( socket.readAll() );
			if ( !buffer.endsWith( C::Separator ) )
			{
				continue;
			}
			if ( buffer.startsWith( "-ERR" ) || buffer.startsWith( "-NOSCRIPT" ) )
			{
				qWarning() << "evaluating redis script error:" << buffer;
				return {};
			}
			int  splitPos;
			auto parts = split( buffer, &splitPos );
			if ( splitPos < buffer.length() )
			{
				continue;
			}
			return( parts );
		}
		else
		{
			QCoreApplication::processEvents();
		}
	}
	return {};
} // Client::evalSha


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Специализация для типа QJsonArray
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
template<>
QJsonArray Client::evalSha< QJsonArray >( const QByteArray& sha, const QVariantList& keys,
                                          const QVariantList& argv )
{
	auto parts = evalSha( sha, keys, argv );
	return( parts.empty() ? QJsonArray {} : QJsonDocument::fromJson( parts.front() ).array() );
}


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Специализация для типа QJsonArray, static вариант
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \param database Идентификатор БД
 * \param host Адрес сервера
 * \param port Номер порта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
template<>
QJsonArray Client::evalSha< QJsonArray >( const QByteArray& sha, const QVariantList& keys,
                                          const QVariantList& argv,
                                          int database, const QString& host, quint16 port )
{
	auto parts = evalSha( sha, keys, argv, database, host, port );
	return( parts.empty() ? QJsonArray {} : QJsonDocument::fromJson( parts.front() ).array() );
}


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Специализация для типа QVariant
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
template<>
QVariant Client::evalSha< QVariant >( const QByteArray& sha, const QVariantList& keys,
                                      const QVariantList& argv )
{
	auto parts = evalSha( sha, keys, argv );
	return( parts.empty() ? QVariant {} : parts.front() );
}


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Специализация для типа QVariantList
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
template<>
QVariantList Client::evalSha< QVariantList >( const QByteArray& sha, const QVariantList& keys,
                                              const QVariantList& argv )
{
	auto parts = evalSha( sha, keys, argv );
	QVariantList data;
	std::copy( parts.begin(), parts.end(), std::back_inserter( data ) );
	return( data );
}


/*!
 * \brief Запускает на выполнение сохраненный на стороне сервера lua скрипт
 *
 * Специализация для типа QVariantHash
 * \param sha SHA сохраненного скрипта
 * \param keys Ключи для выполнения скрипта
 * \param argv Аргументы для выполнения скрипта
 * \return Части результата выполнения скрипта (возвращенного значения)
 */
template<>
QVariantHash Client::evalSha< QVariantHash >( const QByteArray& sha, const QVariantList& keys,
                                              const QVariantList& argv )
{
	auto parts = evalSha( sha, keys, argv );
	QVariantHash data;
	for ( int i = 1; i < parts.size(); i += 2 ) data[parts[i - 1]] = parts[i];
	return( data );
}


/*!
 * \brief Создает сокет для неоткрытой сессии
 * \param Таймаут на установку соединения в мс
 * \return Созданный сокет
 */
QTcpSocket* Client::session( int timeout )
{
	if ( _sessionCount == 0 )
	{
		return( nullptr );
	}
	while ( _sessionCount > _session.size() )
	{
		_session.push_back( new QTcpSocket { this } );
		connect( _session.back(), &QIODevice::readyRead, this, &Client::onReadyRead );
		connect( _session.back(), &QIODevice::bytesWritten, this, &Client::onBytesWritten );
	}
	auto socket = _session[_sessionCount - 1];
	QElapsedTimer timer;
	timer.start();
	while ( socket->state() != QAbstractSocket::ConnectedState )
	{
		if ( connectionFlags() != kConnecting )
		{
			setConnectionFlags( kConnecting );
			emit connectionStateChanged( connectionFlags() );
		}
		if ( timer.elapsed() > timeout )
		{
			setConnectionFlags( kDisconnected );
			emit connectionStateChanged( connectionFlags() );
			return( nullptr );
		}
		if ( socket->state() == QAbstractSocket::UnconnectedState )
		{
			socket->connectToHost( host(), port() );
		}
		QCoreApplication::processEvents();
		if ( ( _database != 0 ) && !select( socket, _database ) )
		{
			setConnectionFlags( kDisconnected );
			emit connectionStateChanged( connectionFlags() );
			return( nullptr );
		}
	}
	auto iter = _outOfDatabase.find( socket );
	if ( iter != _outOfDatabase.end() )
	{
		if ( select( socket, _database ) )
		{
			_outOfDatabase.erase( iter );
		}
		else
		{
			setConnectionFlags( kDisconnected );
			emit connectionStateChanged( connectionFlags() );
			return( nullptr );
		}
	}
	if ( !( isConnected() ) )
	{
		setConnectionFlags( kConnected );
		emit connectionStateChanged( connectionFlags() );
	}
	return( socket );
} // Client::session


/*!
 * \brief Конструктор класса
 * \param count Счетчик открытых сессий
 */
Client::Session::Session( uint& count ) :
	_count{ count }
{
	_count++;
}


/*!
 * \brief Деструктор класса
 */
Client::Session::~Session()
{
	_count--;
}

} // namespace redis

} // namespace myx

#endif // ifndef MYX_REDIS_CLIENT_INL_HPP_
